/*
 * Copyright (c) 2021 Digital Asset (Switzerland) GmbH and/or its affiliates
 *
 * Proprietary code. All rights reserved.
 */

create table event_log
(
    -- Unique name for each instance of the event log (currently can be a string serialized ParticipantId or DomainId)
    name                      NVARCHAR2(300) not null,
    local_offset              NUMBER          not null,
    constraint pk_event_log primary key (name, local_offset),
    -- UTC timestamp of the event in microseconds relative to EPOCH
    ts                        NUMBER          not null,
    -- sequencer counter corresponding to the underlying request, if there is one
    request_sequencer_counter NUMBER,
    -- Optional event ID:
    -- If the event is an Update.TransactionAccepted, this is the transaction id prefixed by `T`, to enable efficient lookup
    -- of the domain that the transaction was executed on.
    -- For timely-rejected transaction (not an Update.TransactionAccepted),
    -- this is the message UUID of the SubmissionRequest, prefixed by `M`, the domain ID, and the separator `#`.
    -- NULL if the event is neither an Update.TransactionAccepted nor was created as part of a timely rejection.
    event_id                  NVARCHAR2(300),
    -- Optional domain ID:
    -- For timely-rejected transactions (not an Update.TransactionAccepted) in the participant event log,
    -- this is the domain ID to which the transaction was supposed to be submitted.
    -- NULL if this is not a timely rejection in the participant event log.
    associated_domain varchar(300),

    -- LedgerSyncEvent serialized using protobuf
    content                   BLOB            not null,
    -- TraceContext is serialized using protobuf
    trace_context             BLOB            not null,
    -- Causality change associated with the event, if there is one
    causality_update               BLOB
);
-- Not strictly required, but sometimes useful.
create index idx_event_log_timestamp on event_log (name, ts);
create unique index idx_event_log_event_id on event_log (event_id);

-- TODO(i4027): merge this into the event_log table once we have enter and leave events
create table transfer_causality_updates (
    -- Name, corresponding to an event_log table
    name NVARCHAR2(300) not null,
    -- Request counter corresponding to the causality update
    request_counter NUMBER not null,
    primary key (name,request_counter),
    -- The request timestamp corresponding to the causality update
    request_timestamp NUMBER not null,
    -- The CausalityUpdate, serialized with protobuf
    causality_update BLOB not null
);

-- Persist a single, linearized, multi-domain event log for the local participant
create table linearized_event_log
(
    -- Global offset
    global_offset         NUMBER GENERATED BY DEFAULT ON NULL AS IDENTITY not null constraint pk_linearized_event_log primary key,
    -- Unique name for each instance of the event log
    name                  NVARCHAR2(300)                                 not null,
    -- Offset in the event log instance designated by name
    local_offset          NUMBER                                          not null,
    constraint foreign_key_event_log foreign key (name, local_offset) references event_log (name, local_offset) on delete cascade,
    -- The participant's local time when the event was published, in microseconds relative to EPOCH.
    -- Increases monotonically with the global offset
    publication_time NUMBER not null
);
create unique index idx_linearized_event_log_offset on linearized_event_log (name, local_offset);
create index idx_linearized_event_log_publication_time on linearized_event_log (publication_time, global_offset);

-- For a party p on a domain d, store the causal dependencies on other domains introduced at a given request counter
create table per_party_causal_dependencies (
    -- Arbitrary primary key
    id                 NUMBER GENERATED BY DEFAULT ON NULL AS IDENTITY not null constraint pk_per_party_causal_dependencies primary key,

    -- The party p
    party_id           NVARCHAR2(300) not null,
    -- The domain of p
    owning_domain_id   NVARCHAR2(300) not null,
    -- The timestamp on owning_domain_id where the causal dependency is introduced
    constraint_ts   NUMBER not null,
    -- The request counter where the causal dependency is introduced, or null if the participant is not connected to owning_domain_id
    request_counter    NUMBER,
    -- If the causal dependency comes from a transfer in, what is the origin domain?
    transfer_origin_domain_if_present NVARCHAR2(300),

    -- The domain for the constraint
    domain_id NVARCHAR2(300) not null,
    -- The timestamp that the p has observed on domain_id
    domain_ts NUMBER not null
);
-- TODO(i7072): Add indices for this table
