/*
 * Copyright (c) 2021 Digital Asset (Switzerland) GmbH and/or its affiliates
 *
 * Proprietary code. All rights reserved.
 */

-- locally computed ACS commitments to a specific period, counter-participant and domain
create table computed_acs_commitments
(
    domain_id                NVARCHAR2(300) not null,
    counter_participant      NVARCHAR2(300) not null,
    -- UTC timestamp in microseconds relative to EPOCH
    from_exclusive           NUMBER          not null,
    -- UTC timestamp in microseconds relative to EPOCH
    to_inclusive             NUMBER          not null,
    -- reconciliation interval in microseconds
    interval_in_microseconds NUMBER          not null,
    -- the "raw" cryptographic commitment (AcsCommitment.CommitmentType) in its serialized format
    commitment               BLOB            not null,
    constraint pk_computed_acs_commitments primary key (counter_participant, domain_id, from_exclusive, to_inclusive),
    constraint check_nonempty_interval_computed check (to_inclusive > from_exclusive)
);

-- ACS commitments received from counter-participants
create table received_acs_commitments
(
    domain_id         NVARCHAR2(300) not null,
    -- the counter-participant who sent the commitment
    sender            NVARCHAR2(300) not null,
    -- UTC timestamp in microseconds relative to EPOCH
    from_exclusive    NUMBER          not null,
    -- UTC timestamp in microseconds relative to EPOCH
    to_inclusive      NUMBER          not null,
    -- the received signed commitment message (including all the other fields) in a serialized form
    signed_commitment BLOB            not null,
    constraint check_to_after_from check (to_inclusive > from_exclusive)
);

create index full_commitment_idx on received_acs_commitments (sender, domain_id, from_exclusive, to_inclusive);

-- the participants whose remote commitments are outstanding
create table outstanding_acs_commitments
(
    domain_id                NVARCHAR2(300) not null,
    -- UTC timestamp in microseconds relative to EPOCH
    from_exclusive           NUMBER          not null,
    -- UTC timestamp in microseconds relative to EPOCH
    to_inclusive             NUMBER          not null,
    -- reconciliation interval in microseconds
    interval_in_microseconds NUMBER          not null,
    counter_participant      NVARCHAR2(300) not null,
    constraint check_nonempty_interval_outstanding check (to_inclusive > from_exclusive)
);

create unique index unique_interval_participant on outstanding_acs_commitments (counter_participant, domain_id, from_exclusive, to_inclusive);

create index outstanding_acs_commitments_by_time on outstanding_acs_commitments (domain_id, from_exclusive);

-- the last timestamp for which the commitments have been computed, stored and sent
create table last_computed_acs_commitments
(
    domain_id NVARCHAR2(300) constraint pk_last_computed_acs_commitments primary key,
    -- UTC timestamp in microseconds relative to EPOCH
    ts        NUMBER not null
);

-- Stores the snapshot ACS commitments (per stakeholder set)
create table commitment_snapshot
(
    domain_id         NVARCHAR2(300) not null,
    -- A stable reference to a stakeholder set, that doesn't rely on the Protobuf encoding being deterministic
    -- a hex-encoded hash (not binary so that hash can be indexed in all db server types)
    -- If this hash requires more than 1000 characters (currently uses 256 due to SHA256), we will need to use extended
    -- character fields (see e.g. https://docs.oracle.com/en/database/oracle/oracle-database/19/spucs/enabling-the-new-extended-data-type-capability.html#GUID-88FB7FFD-4392-49C6-843A-45B49F8A1821)
    stakeholders_hash NVARCHAR2(300) not null,
    stakeholders      BLOB            not null,
    commitment        BLOB            not null,
    constraint pk_commitment_snapshot primary key (stakeholders_hash, domain_id)
);

-- Stores the time (along with a tie-breaker) of the ACS commitment snapshot
create table commitment_snapshot_time
(
    domain_id   NVARCHAR2(300) not null,
    -- UTC timestamp in microseconds relative to EPOCH
    ts          NUMBER          not null,
    tie_breaker NUMBER          not null,
    constraint pk_commitment_snapshot_time primary key (domain_id)
);

-- Remote commitments that were received but could not yet be checked because the local participant is lagging behind
create table commitment_queue
(
    domain_id                NVARCHAR2(300) not null,
    sender                   NVARCHAR2(300) not null,
    counter_participant      NVARCHAR2(300) not null,
    -- UTC timestamp in microseconds relative to EPOCH
    from_exclusive           NUMBER          not null,
    -- UTC timestamp in microseconds relative to EPOCH
    to_inclusive             NUMBER          not null,
    -- reconciliation interval in microseconds
    interval_in_microseconds NUMBER          not null,
    commitment               BLOB            not null,
    commitment_hash          NVARCHAR2(300)  not null, -- A shorter hash (SHA-256) of the commitment for the primary key instead of the BLOB
    constraint check_nonempty_interval_queue check (to_inclusive > from_exclusive),
    constraint pk_commitment_queue primary key (commitment_hash, domain_id, sender, counter_participant, from_exclusive, to_inclusive, interval_in_microseconds)
);

create index commitment_queue_by_time on commitment_queue (domain_id, to_inclusive);
