/*
 * Copyright (c) 2021 Digital Asset (Switzerland) GmbH and/or its affiliates
 *
 * Proprietary code. All rights reserved.
 */

-- we create a unique integer id for each member that is more efficient to use in the events table
-- members can read all events from `registered_ts`
create table sequencer_members
(
    member        NVARCHAR2(300) constraint pk_sequencer_members primary key,
    id            NUMBER GENERATED BY DEFAULT ON NULL AS IDENTITY not null unique,
    registered_ts NUMBER                                          not null,
    -- "1" is true, "0" is false
    enabled CHAR(1) DEFAULT '1' NOT NULL
        CONSTRAINT sequencer_members_enabled_bool check (enabled in ('0','1'))
);

-- stores payloads for send events
-- callers should ensure that a unique id is assigned
create table sequencer_payloads
(
    -- expected to be a timestamp assigned using the partitioned timestamp generator based on the writer index
    id      NUMBER constraint pk_sequencer_payloads primary key,
    -- identifier used by writers to determine on conflicts whether it was the payload they wrote or if another writer
    -- is active with the same writer index (can happen as we use the non-locked unsafe storage method for payloads).
    -- sized to hold a uuid.
    instance_discriminator NVARCHAR2(36) not null,
    content BLOB not null
);

-- sequencer instances maintain the lowest timestamp for which it is safe to read from
create table sequencer_watermarks
(
    node_index integer constraint pk_sequencer_watermarks primary key,
    watermark_ts NUMBER not null,
    -- as suggested by slick for boolean: https://github.com/slick/slick/blob/main/slick/src/main/scala/slick/jdbc/OracleProfile.scala#L318-L324
    -- "1" is true, "0" is false
    sequencer_online CHAR(1) not null
        CONSTRAINT sequencer_watermark_online_bool check (sequencer_online in ('0','1'))
);

-- readers periodically write checkpoints that maps the calculated timer to a timestamp/event-id.
-- when a new subscription is requested from a counter the sequencer can use these checkpoints to find the closest
-- timestamp below the given counter to start the subscription from.
create table sequencer_counter_checkpoints
(
    member  integer not null,
    counter NUMBER  not null,
    ts      NUMBER  not null,
    constraint pk_sequencer_counter_checkpoints primary key (member, counter)
);

-- record the latest acknowledgement sent by a sequencer client for the latest event they have successfully
-- processed and will not re-read.
create table sequencer_acknowledgements
(
    member integer not null,
    instance_id NVARCHAR2(300) not null,
    ts NUMBER not null,
    constraint pk_sequencer_acknowledgements primary key (member, instance_id)
);


-- lower bound of when events can be read
-- if empty it means all events from epoch can be read
-- is updated when sequencer is pruned meaning that earlier events can no longer be read (and likely no longer exist)
create table sequencer_lower_bound (
    single_row_lock char(1) DEFAULT 'X' not null
        CONSTRAINT single_row_constraint check (single_row_lock = 'X') primary key,
    ts NUMBER not null
);

-- lists member instances that have been disabled
-- if a whole member is disabled its enabled flag in the members table will be flipped to false
create table sequencer_disabled_clients (
    member integer not null,
    instance_id NVARCHAR2(300) not null,
    constraint pk_sequencer_disabled_clients primary key (member, instance_id)
)
