/*
 * Copyright (c) 2021 Digital Asset (Switzerland) GmbH and/or its affiliates
 *
 * Proprietary code. All rights reserved.
 */

-- Stores the immutable contracts, however a creation of a contract can be rolled back.
create table contracts
(
    -- As a participant can be connected to multiple domains, the transactions are stored under a domain id.
    -- Only setting this to 500 characters, so we can create an index along with package_id and template_id
    domain_id               NUMBER not null,
    contract_id             NVARCHAR2(300) not null,
    -- The contract is serialized using the LF contract proto serializer.
    instance                BLOB            not null,
    -- Metadata: signatories, stakeholders, keys
    -- Stored as a Protobuf blob as H2 will only support typed arrays in 1.4.201
    -- TODO(#3256): change when H2 is upgraded
    metadata                BLOB            not null,
    -- The ledger time when the contract was created.
    ledger_create_time      NVARCHAR2(300) not null,
    -- The request counter of the request that created or divulged the contract
    request_counter         NUMBER not null,
    -- The transaction that created the contract; null for divulged contracts
    creating_transaction_id BLOB,
    -- We store metadata of the contract instance for inspection
    -- We assume that it's safe to store package_id as a length-limited string as it is currently a SHA256-hash
    -- and this is unlikely to ever change
    -- Only setting this to 300 characters, so we can create an index along with package_id and template_id
    package_id              NVARCHAR2(300) not null,
    -- While this field is technically unlimited, the upstream index DB uses 4000 chars for template_id and therefore so do we
    -- https://github.com/digital-asset/daml/blob/a205d0e44631679c958c53873df3a5a1513464c9/ledger/participant-integration-api/src/main/resources/db/migration/oracle/V1__Init.sql#L216
    template_id             VARCHAR2(4000) not null,
    constraint pk_contracts primary key (contract_id, domain_id)
);

-- Index to speedup ContractStore.find
-- domain_id comes first, because there is always a constraint on it.
-- package_id comes before template_id, because queries with package_id and without template_id make more sense than vice versa.
-- contract_id is left out, because a query with domain_id and contract_id can be served with the primary key.
create index idx_contracts_find on contracts(template_id, domain_id, package_id);
